import hashlib
from datetime import datetime
from typing import Optional

from cpl_core.configuration import ConfigurationABC
from cpl_core.database.context import DatabaseContextABC
from cpl_core.logging import LoggerABC
from cpl_discord.service import DiscordBotServiceABC
from flask import request

from bot_api.configuration.authentication_settings import AuthenticationSettings
from bot_core.configuration.feature_flags_enum import FeatureFlagsEnum
from bot_core.configuration.feature_flags_settings import FeatureFlagsSettings
from bot_data.abc.api_key_repository_abc import ApiKeyRepositoryABC
from bot_data.abc.game_server_repository_abc import GameServerRepositoryABC
from bot_data.abc.server_repository_abc import ServerRepositoryABC
from bot_data.abc.user_game_ident_repository_abc import UserGameIdentRepositoryABC
from bot_data.abc.user_joined_game_server_repository_abc import UserJoinedGameServerRepositoryABC
from bot_data.abc.user_repository_abc import UserRepositoryABC
from bot_data.model.api_key import ApiKey
from bot_data.model.server_config import ServerConfig
from bot_data.model.user_joined_game_server import UserJoinedGameServer
from bot_data.model.user_role_enum import UserRoleEnum
from bot_graphql.abc.query_abc import QueryABC
from modules.permission.service.permission_service import PermissionService


class UserJoinedGameServerMutation(QueryABC):
    def __init__(
        self,
        config: ConfigurationABC,
        logger: LoggerABC,
        servers: ServerRepositoryABC,
        users: UserRepositoryABC,
        user_joined_game_servers: UserJoinedGameServerRepositoryABC,
        api_keys: ApiKeyRepositoryABC,
        game_servers: GameServerRepositoryABC,
        user_game_idents: UserGameIdentRepositoryABC,
        bot: DiscordBotServiceABC,
        db: DatabaseContextABC,
        permissions: PermissionService,
        auth_settings: AuthenticationSettings,
    ):
        QueryABC.__init__(self, "UserJoinedGameServerMutation")

        self._config = config
        self._logger = logger
        self._servers = servers
        self._users = users
        self._user_joined_game_servers = user_joined_game_servers
        self._api_keys = api_keys
        self._game_servers = game_servers
        self._user_game_idents = user_game_idents
        self._bot = bot
        self._db = db
        self._permissions = permissions
        self._auth_settings = auth_settings

        self.set_field("userJoined", self.resolve_user_joined)
        self.set_field("userLeft", self.resolve_user_left)

    def _get_api_key_str(self, api_key: ApiKey) -> str:
        return hashlib.sha256(
            f"{api_key.identifier}:{api_key.key}+{self._auth_settings.secret_key}".encode("utf-8")
        ).hexdigest()

    def _get_api_key(self) -> Optional[ApiKey]:
        authorization = request.headers.get("Authorization").split()
        if len(authorization) != 2:
            return None

        api_key_str = authorization[1]
        return (
            self._api_keys.get_api_keys()
            .where(lambda key: self._get_api_key_str(key) == api_key_str)
            .first_or_default()
        )

    def resolve_user_joined(self, *_, input: dict):
        game_ident = self._user_game_idents.get_user_game_ident_by_ident(input["ident"])
        user = game_ident.user

        settings: ServerConfig = self._config.get_configuration(f"ServerConfig_{user.server.discord_id}")
        if not FeatureFlagsSettings.get_flag_from_dict(settings.feature_flags, FeatureFlagsEnum.game_server):
            raise Exception("Feature not activated")

        self._can_user_mutate_data(user.server, UserRoleEnum.admin)

        active = self._user_joined_game_servers.find_active_user_joined_game_server_by_user_id(user.id)
        if active is not None:
            self._logger.debug(
                __name__,
                f"Skip UserJoinedGameServer for user {user.id}. User already plays on {active.game_server}.",
            )
            return

        new = UserJoinedGameServer(user, game_ident.game_server, datetime.now())
        self._user_joined_game_servers.add_user_joined_game_server(new)
        self._db.save_changes()

        return self._user_joined_game_servers.get_active_user_joined_game_server_by_user_id(user.id)

    def resolve_user_left(self, *_, input: dict):
        game_ident = self._user_game_idents.find_user_game_ident_by_ident(input["ident"])
        if game_ident is None:
            return
        user = game_ident.user

        settings: ServerConfig = self._config.get_configuration(f"ServerConfig_{user.server.discord_id}")
        if not FeatureFlagsSettings.get_flag_from_dict(settings.feature_flags, FeatureFlagsEnum.game_server):
            raise Exception("Feature not activated")

        self._can_user_mutate_data(user.server, UserRoleEnum.admin)

        active = self._user_joined_game_servers.find_active_user_joined_game_server_by_user_id(user.id)
        if active is None:
            return
        active.leaved_on = datetime.now()

        settings: ServerConfig = self._config.get_configuration(f"ServerConfig_{user.server.discord_id}")

        ontime = round((active.leaved_on - active.joined_on).total_seconds() / 3600, 2)
        old_xp = user.xp
        user.xp += round(ontime * settings.xp_per_ontime_hour)

        self._user_joined_game_servers.update_user_joined_game_server(active)
        self._users.update_user(user)
        self._db.save_changes()
        self._logger.debug(
            __name__,
            f"User {user} leaved_on {active.leaved_on}. Ontime: {ontime}h | xp: from {old_xp} to {user.xp}",
        )

        return active
